"""Property wrappers handling potentially obsolete formats..."""
from aomi.helpers import warning
from aomi.vault import app_id_name


def app_id_itself(app_obj, data):
    """Determines the application ID to use"""
    app_id = None
    if 'app_id' in data:
        warning('Defining app_id within the app yaml is deprecated')
        app_id = data['app_id']
    elif 'app_id' in app_obj:
        app_id = app_obj['app_id']
    else:
        app_id = app_id_name(app_obj)

    return app_id


def aws_region(secret, aws_obj):
    """Return the AWS region with appropriate output"""
    if 'region' in secret:
        return secret['region']
    else:
        # see https://github.com/Autodesk/aomi/issues/40
        warning('Defining region in the AWS yaml is deprecated')
        return aws_obj['region']


def aws_roles(secret, aws_obj):
    """Return the AWS roles with appropriate output"""
    if 'roles' in secret:
        return secret['roles']
    else:
        # see https://github.com/Autodesk/aomi/issues/40
        warning('Defining roles within the AWS yaml is deprecated')
        return aws_obj['roles']


def app_id_policy_file(app_obj, data):
    """Determines the correct policy file name, checking both the
    proper and legacy location"""
    policy_file = None
    if 'policy' in data:
        warning('Defining policy_name within the app yaml is deprecated')
        policy_file = data['policy']
    elif 'policy' in app_obj:
        policy_file = app_obj['policy']

    return policy_file


def app_id_policy_name(app_obj, data):
    """Determines the policy name, checking both the proper
    and the legacy location"""
    policy_name = None
    if 'policy_name' in data:
        warning('Defining policy_name within the app yaml is deprecated')
        policy_name = data['policy_name']
    elif 'policy_name' in data:
        policy_name = app_obj['policy_name']
    else:
        policy_name = app_id_name(app_obj)

    return policy_name



""" Vault interactions """
from __future__ import print_function
import os
import socket
import hvac
import yaml
# need to override those SSL warnings
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from aomi.helpers import log, cli_hash, merge_dicts, abspath
from aomi.template import render, load_var_files
from aomi.error import output as error_output
import aomi.error
import aomi.exceptions

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def approle_token(vault_client, role_id, secret_id):
    """Returns a vault token based on the role and seret id"""
    resp = vault_client.auth_approle(role_id, secret_id)
    if 'auth' in resp and 'client_token' in resp['auth']:
        return resp['auth']['client_token']
    else:
        raise aomi.exceptions.AomiCredentials('invalid approle')


def app_token(vault_client, app_id, user_id):
    """Returns a vault token based on the app and user id."""
    resp = vault_client.auth_app_id(app_id, user_id)
    if 'auth' in resp and 'client_token' in resp['auth']:
        return resp['auth']['client_token']
    else:
        raise aomi.exceptions.AomiCredentials('invalid apptoken')


def initial_token(vault_client, opt):
    """Generate our first token based on workstation configuration"""
    home = os.environ['HOME'] if 'HOME' in os.environ else \
        os.environ['USERPROFILE']

    token_file = os.environ.get('VAULT_TOKEN_FILE',
                                os.path.join(home, ".vault-token"))
    app_file = os.environ.get('AOMI_APP_FILE',
                              os.path.join(home, ".aomi-app-token"))
    token_file = abspath(token_file)
    app_file = abspath(app_file)
    if 'VAULT_TOKEN' in os.environ and len(os.environ['VAULT_TOKEN']) > 0:
        log('Token derived from VAULT_TOKEN environment variable', opt)
        return os.environ['VAULT_TOKEN'].strip()
    elif 'VAULT_USER_ID' in os.environ and \
         'VAULT_APP_ID' in os.environ and \
         len(os.environ['VAULT_USER_ID']) > 0 and \
         len(os.environ['VAULT_APP_ID']) > 0:
        token = app_token(vault_client,
                          os.environ['VAULT_APP_ID'].strip(),
                          os.environ['VAULT_USER_ID'].strip())
        log("Token derived from VAULT_APP_ID and VAULT_USER_ID", opt)
        return token
    elif 'VAULT_ROLE_ID' in os.environ and \
         'VAULT_SECRET_ID' in os.environ and \
         len(os.environ['VAULT_ROLE_ID']) > 0 and \
         len(os.environ['VAULT_SECRET_ID']) > 0:
        token = approle_token(vault_client,
                              os.environ['VAULT_ROLE_ID'],
                              os.environ['VAULT_SECRET_ID'])
        log("Token derived from VAULT_ROLE_ID and VAULT_SECRET_ID", opt)
        return token
    elif os.path.exists(app_file):
        token = yaml.safe_load(open(app_file).read().strip())
        if 'app_id' in token and 'user_id' in token:
            token = app_token(vault_client,
                              token['app_id'],
                              token['user_id'])
            log("Token derived from %s" % app_file, opt)
            return token
    elif os.path.exists(token_file):
        log("Token derived from %s" % token_file, opt)
        return open(token_file, 'r').read().strip()
    else:
        raise aomi.exceptions.AomiCredentials('unknown method')


def token_meta(operation, opt):
    """Generates metadata for a token"""
    meta = {
        'via': 'aomi',
        'operation': operation,
        'hostname': socket.gethostname()
    }
    if 'USER' in os.environ:
        meta['unix_user'] = os.environ['USER']

    if opt.metadata:
        meta_bits = opt.metadata.split(',')
        for meta_bit in meta_bits:
            key, value = meta_bit.split('=')

        if key not in meta:
            meta[key] = value

    for key, value in meta.items():
        log("Token metadata %s %s" % (key, value), opt)

    return meta


def operational_token(vault_client, operation, opt):
    """Return a properly annotated token for our use."""
    display_name = vault_client.lookup_token()['data']['display_name']
    args = {
        'lease': opt.lease,
        'display_name': display_name,
        'meta': token_meta(operation, opt)
    }
    try:
        token = vault_client.create_token(**args)
    except (hvac.exceptions.InvalidRequest,
            hvac.exceptions.Forbidden) as vault_exception:
        if vault_exception.errors[0] == 'permission denied':
            error_output("Permission denied creating operational token", opt)
        else:
            raise vault_exception

    log("Using lease of %s" % opt.lease, opt)
    return token['auth']['client_token']


def client(operation, opt):
    """Return a vault client"""
    if 'VAULT_ADDR' not in os.environ:
        raise aomi.exceptions.AomiError('VAULT_ADDR must be defined')

    vault_host = os.environ['VAULT_ADDR']

    ssl_verify = True
    if 'VAULT_SKIP_VERIFY' in os.environ:
        if os.environ['VAULT_SKIP_VERIFY'] == '1':
            log('Skipping SSL Validation!', opt)
            ssl_verify = False

    log("Connecting to %s" % vault_host, opt)
    vault_client = hvac.Client(vault_host, verify=ssl_verify)
    vault_client.token = initial_token(vault_client, opt)
    if not vault_client.is_authenticated():
        raise aomi.exceptions.AomiCredentials('initial token')

    vault_client.token = operational_token(vault_client, operation, opt)
    if not vault_client.is_authenticated():
        raise aomi.exceptions.AomiCredentials('operational token')

    return vault_client


def get_secretfile(opt):
    """Renders, YAMLs, and returns the Secretfile construct"""
    secretfile_path = abspath(opt.secretfile)
    obj = merge_dicts(load_var_files(opt),
                      cli_hash(opt.extra_vars))
    return yaml.safe_load(render(secretfile_path, obj))


def app_id_name(app_obj):
    """Determines the proper app id name"""
    name = None
    if 'name' in app_obj:
        name = app_obj['name']
    else:
        name = os.path.splitext(os.path.basename(app_obj['app_file']))[0]

    return name



"""Handles the various kinds of secret seeding which we do"""
import re
from uuid import uuid4
import yaml
import hvac
import aomi.exceptions
from aomi.helpers import hard_path, log, \
    warning, merge_dicts, cli_hash, random_word
import aomi.validation
from aomi.error import output as error_output
from aomi.validation import sanitize_mount
from aomi.template import render, load_var_files
from aomi.vault import app_id_name
from aomi.util import validate_entry
import aomi.legacy


def unmount(client, backend, path):
    """Unmount a given mountpoint"""
    backends = client.list_secret_backends()
    if is_mounted(path, backends, backend):
        client.disable_secret_backend(path)


def is_mounted(mount, backends, style):
    """Determine whether a backend of a certain type is mounted"""
    for mount_name, values in backends.items():
        b_norm = '/'.join([x for x in mount_name.split('/') if x])
        m_norm = '/'.join([x for x in mount.split('/') if x])
        if (m_norm == b_norm) and values['type'] == style:
            return True

    return False


def maybe_mount(client, backend, mount, opt):
    """Will ensure a mountpoint exists, or bail with a polite error"""
    backends = client.list_secret_backends()
    if not is_mounted(mount, backends, backend):
        log("Specifying a inline mountpoint is deprecated", opt)
        actually_mount(client, backend, mount)


def actually_mount(client, backend, mount):
    """Actually mount something in Vault"""
    try:
        client.enable_secret_backend(backend, mount_point=mount)
    except hvac.exceptions.InvalidRequest as exception:
        client.revoke_self_token()
        match = re.match('existing mount at (?P<path>.+)', str(exception))
        if match:
            e_msg = "%s has a mountpoint conflict with %s" % \
                    (mount, match.group('path'))
            raise aomi.exceptions.VaultConstraint(e_msg)
        else:
            raise exception


def is_auth_backend(client, backend):
    """Checks to see if an auth backend exists yet"""
    backends = client.list_auth_backends().keys()
    backends = [x.rstrip('/') for x in backends]
    return backend in backends


def ensure_auth(client, auth):
    """Will ensure a particular auth endpoint is mounted"""
    if not is_auth_backend(client, auth):
        client.enable_auth_backend(auth)


def write(client, path, varz, opt):
    """Write to Vault while handling non-surprising errors."""
    try:
        client.write(path, **varz)
    except (hvac.exceptions.InvalidRequest,
            hvac.exceptions.Forbidden) as vault_exception:
        client.revoke_self_token()
        if vault_exception.errors[0] == 'permission denied':
            error_output("Permission denied writing to %s" % path, opt)
        else:
            raise vault_exception


def delete(client, path, opt):
    """Delete from Vault while handling non-surprising errors."""
    try:
        client.delete(path)
    except (hvac.exceptions.InvalidRequest,
            hvac.exceptions.Forbidden) as vault_exception:
        client.revoke_self_token()
        if vault_exception.errors[0] == 'permission denied':
            error_output("Permission denied deleting %s" % path, opt)
        else:
            raise vault_exception


def var_file(client, secret, opt):
    """Seed a var_file into Vault"""
    aomi.validation.var_file_obj(secret)
    my_mount = sanitize_mount(secret['mount'])
    path = "%s/%s" % (my_mount, secret['path'])
    if not validate_entry(secret, path, opt):
        return

    var_file_name = hard_path(secret['var_file'], opt.secrets)
    aomi.validation.secret_file(var_file_name)
    varz = yaml.safe_load(open(var_file_name).read())

    maybe_mount(client, 'generic', my_mount, opt)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    if secret.get('state', 'present') == 'present':
        write(client, path, varz, opt)
        log('wrote var_file %s into %s' % (
            var_file_name,
            path), opt)
    else:
        delete(client, path, opt)
        log('deleted var_file %s from %s' % (
            var_file_name, path), opt)


def aws(client, secret, opt):
    """Seed an aws_file into Vault"""
    aomi.validation.aws_file_obj(secret)

    my_mount = sanitize_mount(secret['mount'])
    aws_path = "%s/config/root" % my_mount
    if not validate_entry(secret, aws_path, opt):
        return

    if secret.get('state', 'present') == 'absent':
        unmount(client, 'aws', my_mount)
        log("Unmounted AWS %s" % aws_path, opt)
        return
    else:
        maybe_mount(client, 'aws', my_mount, opt)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    aws_file_path = hard_path(secret['aws_file'], opt.secrets)
    aomi.validation.secret_file(aws_file_path)

    aws_obj = yaml.safe_load(open(aws_file_path, 'r').read())
    aomi.validation.aws_secret_obj(aws_file_path, aws_obj)

    region = aomi.legacy.aws_region(secret, aws_obj)
    if region is None:
        client.revoke_self_token()
        raise aomi.exceptions.AomiData('missing aws region')

    roles = aomi.legacy.aws_roles(secret, aws_obj)
    if roles is None:
        client.revoke_self_token()
        raise aomi.exceptions.AomiData('missing aws roles')

    obj = {
        'access_key': aws_obj['access_key_id'],
        'secret_key': aws_obj['secret_access_key'],
        'region': region
    }
    write(client, aws_path, obj, opt)
    log('wrote aws secrets %s into %s' % (
        aws_file_path,
        aws_path), opt)

    ttl_obj, lease_msg = grok_ttl(secret, aws_obj)
    if ttl_obj:
        write(client,
              "%s/config/lease" % (secret['mount']),
              ttl_obj,
              opt)
        log("Updated lease for %s %s" % (secret['mount'], lease_msg), opt)

    seed_aws_roles(client, secret['mount'], roles, opt)


def grok_ttl(secret, aws_obj):
    """Parses the TTL information, keeping in mind old format"""
    ttl_obj = {}
    lease_msg = ''
    if 'lease' in secret:
        ttl_obj['lease'] = secret['lease']
        lease_msg = "lease:%s" % (ttl_obj['lease'])

    if 'lease_max' in secret:
        ttl_obj['lease_max'] = secret['lease_max']
    else:
        if 'lease' in ttl_obj:
            ttl_obj['lease_max'] = ttl_obj['lease']

    if lease_msg == '':
        if 'lease' in aws_obj:
            ttl_obj['lease'] = aws_obj['lease']
            lease_msg = "lease:%s" % (ttl_obj['lease'])

        if 'lease_max' in aws_obj:
            ttl_obj['lease_max'] = aws_obj['lease_max']
        else:
            if 'lease' in ttl_obj:
                ttl_obj['lease_max'] = ttl_obj['lease']

        if lease_msg != '':
            # see https://github.com/Autodesk/aomi/issues/40
            warning('Setting lease and lease_max from the '
                    'AWS yaml is deprecated')

    if 'lease_max' in ttl_obj:
        lease_msg = "%s lease_max:%s" % (lease_msg, ttl_obj['lease_max'])

    return ttl_obj, lease_msg


def seed_aws_roles(client, mount, roles, opt):
    """Handles the seeding of roles associated with an AWS account"""
    for role in roles:
        aomi.validation.aws_role_obj(role)

        role_path = "%s/roles/%s" % (mount, role['name'])
        if role.get('state', 'present') == 'present':
            if 'policy' in role:
                role_file = hard_path(role['policy'], opt.policies)
                role_template_obj = role.get('vars', {})
                cli_obj = merge_dicts(load_var_files(opt),
                                      cli_hash(opt.extra_vars))
                obj = merge_dicts(role_template_obj, cli_obj)
                data = render(role_file, obj)
                log('writing inline role %s from %s' %
                    (role['name'], role_file), opt)
                write(client, role_path, {'policy': data}, opt)
            elif 'arn' in role:
                log('writing role %s for %s' %
                    (role['name'], role['arn']), opt)
                write(client, role_path, {'arn': role['arn']}, opt)
        else:
            log('removing role %s' % role['name'], opt)
            delete(client, role_path, opt)


def app_users(client, app_id, p_users, opt):
    """Write out users for an application"""
    for user in p_users:
        if 'id' not in user:
            client.revoke_self_token()
            raise aomi.exceptions.AomiData("Invalid user definition %s" % user)

        user_path = "auth/app-id/map/user-id/%s" % user['id']
        if user.get('state', 'present') == 'absent':
            delete(client, user_path, opt)
        else:
            user_obj = {'value': app_id}
            if 'cidr' in user:
                user_obj['cidr_block'] = user['cidr']

            write(client, user_path, user_obj, opt)


def app_policy(client, policy_name, policy_file, app_obj, opt):
    """Ensures the policy portion of an app is correct"""

    if policy_file:
        p_data = policy_data(policy_file, app_obj.get('policy_vars', {}), opt)
        if policy_name in client.list_policies():
            if p_data != client.get_policy(policy_name):
                client.revoke_self_token()
                e_msg = "Policy %s already exists " \
                        "and content differs" % policy_name
                raise aomi.exceptions.AomiData(e_msg)

        write_policy(policy_name, p_data, client, opt)
    else:
        if policy_name not in client.list_policies():
            client.revoke_self_token()
            e_msg = "Policy %s is not inline but does not exist" % policy_name
            raise aomi.exceptions.AomiData(e_msg)

        log("Using existing policy %s" % policy_name, opt)


def app(client, app_obj, opt):
    """Seed an app file into Vault"""
    if 'app_file' not in app_obj:
        client.revoke_self_token()
        raise aomi.exceptions.AomiData("Invalid app definition %s" % app_obj)

    name = app_id_name(app_obj)
    if not validate_entry(app_obj, "app-id/%s" % name, opt):
        return

    app_file = hard_path(app_obj['app_file'], opt.secrets)
    aomi.validation.secret_file(app_file)
    data = yaml.safe_load(open(app_file).read())
    app_id = aomi.legacy.app_id_itself(app_obj, data)
    app_path = "auth/app-id/map/app-id/%s" % app_id

    if app_obj.get('state', 'present') == 'absent':
        delete(client, app_path, opt)
    else:
        ensure_auth(client, 'app-id')
        if opt.mount_only:
            log("Only enabling app-id", opt)
            return

        if 'users' not in data:
            client.revoke_self_token()
            raise aomi.exceptions.AomiData("Invalid app file %s" % app_file)

        policy_name = aomi.legacy.app_id_policy_name(app_obj, data)
        policy_file = aomi.legacy.app_id_policy_file(app_obj, data)
        app_policy(client, policy_name, policy_file, app_obj, opt)

        app_obj = {'value': policy_name, 'display_name': name}
        write(client, app_path, app_obj, opt)
        r_users = data.get('users', [])
        app_users(client, app_id, r_users, opt)
        log('created %d users in application %s' % (len(r_users), name), opt)


def files(client, secret, opt):
    """Seed files into Vault"""
    aomi.validation.file_obj(secret)
    obj = {}
    my_mount = sanitize_mount(secret['mount'])
    vault_path = "%s/%s" % (my_mount, secret['path'])
    if not validate_entry(secret, vault_path, opt):
        return

    maybe_mount(client, 'generic', my_mount, opt)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    if secret.get('state', 'present') == 'present':
        for sfile in secret.get('files', []):
            if 'source' not in sfile or 'name' not in sfile:
                client.revoke_self_token()
                e_msg = "Invalid file specification %s" % sfile
                raise aomi.exceptions.AomiData(e_msg)

            filename = hard_path(sfile['source'], opt.secrets)
            aomi.validation.secret_file(filename)
            data = open(filename, 'r').read()
            obj[sfile['name']] = data
            log('writing file %s into %s/%s' % (
                filename,
                vault_path,
                sfile['name']), opt)

        write(client, vault_path, obj, opt)
    else:
        rmfiles = ','.join([f['source'] for f in secret.get('files', [])])
        log("Removing files %s from %s" % (
            rmfiles, vault_path), opt)
        delete(client, vault_path, opt)


def policy_data(file_name, policy_vars, opt):
    """Returns the rendered policy"""
    policy_path = hard_path(file_name, opt.policies)
    cli_obj = merge_dicts(load_var_files(opt),
                          cli_hash(opt.extra_vars))
    obj = merge_dicts(policy_vars, cli_obj)
    return render(policy_path, obj)


def write_policy(policy_name, data, client, opt):
    """Actually write a policy to vault (renders as a template first)"""
    log('writing policy %s' % policy_name, opt)
    client.set_policy(policy_name, data)


def policy(client, secret, opt):
    """Seed a standalone policy into Vault"""
    aomi.validation.policy_obj(secret)

    policy_name = secret['name']
    if not validate_entry(secret, "policy/%s" % policy_name, opt):
        return

    if secret.get('state', 'present') == 'present':
        data = policy_data(secret['file'], secret.get('vars', {}), opt)
        write_policy(policy_name, data, client, opt)
    else:
        log('removing policy %s' % policy_name, opt)
        client.delete_policy(policy_name)


def audit_logs(client, log_obj, opt):
    """Creates an audit log entry in Vault"""
    aomi.validation.audit_log_obj(log_obj)
    log_type = log_obj['type']
    vault_path = log_obj.get('path', log_type)
    if not aomi.validation.tag_check(log_obj, vault_path, opt):
        return

    existing_backends = client.list_audit_backends()['data']
    audit_path = "%s/" % vault_path
    if audit_path in existing_backends.keys():
        if log_type == existing_backends[audit_path]['type']:
            log("audit log %s already exists" % vault_path, opt)
            return
        else:
            client.revoke_self_token()
            e_msg = "conflicting audit log at %s" % vault_path
            raise aomi.exceptions.VaultConstraint(e_msg)

    obj = {
        'type': log_obj['type']
    }
    if log_type == 'file':
        obj['file_path'] = log_obj['file_path']

    if log_type == 'syslog':
        if 'tag' in log_obj:
            obj['tag'] = log_obj['tag']

        if 'facility' in log_obj:
            obj['facility'] = log_obj['facility']

    if 'description' in log_obj:
        client.enable_audit_backend(log_type,
                                    description=log_obj['description'],
                                    options=obj,
                                    name=vault_path)
    else:
        client.enable_audit_backend(log_type,
                                    options=obj,
                                    name=vault_path)

    log("created %s audit log at %s" % (log_type, vault_path), opt)


def users(client, user_obj, opt):
    """Creates userpass users in Vault"""
    aomi.validation.user_obj(user_obj)

    name = user_obj['username']

    if not aomi.validation.tag_check(user_obj,
                                     "userpass/%s" % name,
                                     opt):
        return

    password_file = hard_path(user_obj['password_file'],
                              opt.secrets)
    aomi.validation.secret_file(password_file)
    password = open(password_file).readline().strip()

    ensure_auth(client, 'userpass')

    user_path = "auth/userpass/users/%s" % name
    if user_obj.get('state', 'present') == 'present':
        v_obj = {
            'password': password,
            'policies': ','.join(user_obj['policies'])
        }
        write(client, user_path, v_obj, opt)
    else:
        delete(client, user_path, opt)


def approle(client, approle_obj, opt):
    """Will seed application role information into a Vault"""
    aomi.validation.approle_obj(approle_obj)
    name = approle_obj['name']
    if not aomi.validation.tag_check(approle_obj,
                                     "approle/%s" % name,
                                     opt):
        return

    ensure_auth(client, 'approle')

    policies = approle_obj['policies']

    role_obj = {
        'policies': ','.join(policies)
    }

    if 'cidr_list' in approle_obj:
        role_obj['bound_cidr_list'] = ','.join(approle_obj['cidr_list'])
    else:
        role_obj['bound_cidr_list'] = ''

    if 'secret_uses' in approle_obj:
        role_obj['secret_id_num_uses'] = approle_obj['secret_uses']

    if 'secret_ttl' in approle_obj:
        role_obj['secret_id_ttl'] = approle_obj['secret_ttl']

    log("creating approle %s" % name, opt)
    client.create_role(name, **role_obj)


def generated_key(key, opt):
    """Create the proper generated key value"""
    key_name = key['name']
    if key['method'] == 'uuid':
        log("Setting %s to a uuid" % key_name, opt)
        return str(uuid4())
    elif key['method'] == 'words':
        log("Setting %s to random words" % key_name, opt)
        return random_word()
    elif key['method'] == 'static':
        if 'value' not in key.keys():
            raise aomi.exceptions.AomiData("Missing static value")

        log("Setting %s to a static value" % key_name, opt)
        return key['value']
    else:
        raise aomi.exceptions.AomiData("Unexpected generated secret method %s"
                                       % key['method'])


def generated(client, obj, opt):
    """Will provision some random strings into vault, as requested"""
    aomi.validation.generated_obj(obj)
    my_mount = sanitize_mount(obj['mount'])
    vault_path = "%s/%s" % (my_mount, obj['path'])
    if not aomi.validation.tag_check(obj, vault_path, opt):
        return

    maybe_mount(client, 'generic', my_mount, opt)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    existing = {}
    resp = client.read(vault_path)
    if resp:
        existing = resp['data']

    if obj.get('state', 'present') == 'present':
        secret_obj = {}
        for key in obj['keys']:
            key_name = key['name']
            if key_name in existing and not key.get('overwrite'):
                log("Not overwriting %s/%s" % (vault_path, key_name), opt)
                continue

            secret_obj[key_name] = generated_key(key, opt)

        genseclen = len(secret_obj.keys())
        if genseclen > 0:
            update_msg = "Writing %s generated secrets to %s" % \
                         (genseclen, vault_path)
            log(update_msg, opt)
            write(client, vault_path, secret_obj, opt)
    else:
        log("Removing generated secret at %s" % vault_path, opt)
        delete(client, vault_path, opt)


def mount_path(client, obj, opt):
    """Manage a Vault mountpoint"""
    aomi.validation.mount_obj(obj)
    path = obj['path']
    if not validate_entry(obj, path, opt):
        return

    backends = client.list_secret_backends()
    mounted = is_mounted(path, backends, 'generic')
    if obj.get('state', 'present') == 'present':
        if not mounted:
            actually_mount(client, 'generic', path)
            log("Mounted %s" % (path), opt)
    else:
        if mounted:
            unmount(client, 'generic', path)
            log("Mounted %s" % (path), opt)


def duo_enable(client, backend, opt):
    """Set the MFA type to DUO"""
    obj = {
        'type': 'duo'
    }
    path = "auth/%s/mfa_config" % backend
    existing = client.read(path)
    if existing \
       and 'data' in existing \
       and 'type' in existing['data'] \
       and existing['data']['type'] == 'duo':
        log("Auth backend %s already configured for DUO" % backend, opt)
    else:
        write(client, path, obj, opt)
        log("Auth backend %s now configured for DUO" % backend, opt)


def duo_access(client, obj, opt):
    """Sets the DUO access configuration"""
    creds_file_name = hard_path(obj['creds'], opt.secrets)
    aomi.validation.secret_file(creds_file_name)
    creds = yaml.safe_load(open(creds_file_name).read())
    duo_obj = {
        'ikey': creds['key'],
        'skey': creds['secret'],
        'host': obj['host']
    }
    path = "auth/%s/duo/access" % obj['backend']
    write(client, path, duo_obj, opt)


def duo(client, obj, opt):
    """Manage DUO MFA on Auth endpoints"""
    aomi.validation.duo_obj(obj)
    backend = obj['backend']
    if not aomi.validation.tag_check(obj,
                                     "duo/%s" % backend,
                                     opt):
        return

    if obj.get('state', 'present') == 'present':
        if not is_auth_backend(client, backend):
            e_msg = "Backend %s not found when configuring duo" % backend
            aomi.exceptions.VaultConstraint(e_msg)

        duo_enable(client, backend, opt)
        duo_access(client, obj, opt)
    else:
        if not is_auth_backend(client, backend):
            return

        write(client, "auth/%s/mfa_config" % backend, {}, opt)
        write(client, "auth/%s/duo/access" % backend, {}, opt)
        log("Removed DUO MFA support from %s" % backend, opt)



""" The aomi "seed" loop """
from __future__ import print_function
import os
from shutil import rmtree
import tempfile
# need to override those SSL warnings
import aomi.seed
from aomi.filez import thaw
from aomi.vault import get_secretfile
import aomi.error
import aomi.exceptions


def seed_secrets(config, vault_client, opt):
    """Seed our various secrets"""
    for secret in config.get('secrets', []):
        if 'var_file' in secret:
            aomi.seed.var_file(vault_client, secret, opt)
        elif 'aws_file' in secret:
            aomi.seed.aws(vault_client, secret, opt)
        elif 'files' in secret:
            aomi.seed.files(vault_client, secret, opt)
        elif 'generated' in secret:
            aomi.seed.generated(vault_client, secret['generated'], opt)
        else:
            raise aomi.exceptions.AomiData("secret element %s" % secret)


def auto_thaw(opt):
    """Will thaw into a temporary location"""
    icefile = opt.thaw_from
    if not os.path.exists(icefile):
        raise aomi.exceptions.IceFile("%s missing" % icefile)

    thaw(icefile, opt)
    return opt


def seed(vault_client, opt):
    """Will provision vault based on the definition within a Secretfile"""
    if opt.thaw_from:
        opt.secrets = tempfile.mkdtemp('aomi-thaw')
        auto_thaw(opt)

    config = get_secretfile(opt)
    seed_secrets(config, vault_client, opt)

    for policy in config.get('policies', []):
        aomi.seed.policy(vault_client, policy, opt)

    for app in config.get('apps', []):
        aomi.seed.app(vault_client, app, opt)

    for user in config.get('users', []):
        aomi.seed.users(vault_client, user, opt)

    for audit_log in config.get('audit_logs', []):
        aomi.seed.audit_logs(vault_client, audit_log, opt)

    for approle in config.get('approles', []):
        aomi.seed.approle(vault_client, approle, opt)

    for mount in config.get('mounts', []):
        aomi.seed.mount_path(vault_client, mount, opt)

    for duo in config.get('duo', []):
        aomi.seed.duo(vault_client, duo, opt)

    if opt.thaw_from:
        rmtree(opt.secrets)



""" Helpers for aomi that are used throughout the application """
from __future__ import print_function
import collections
import itertools as IT
import sys
import os
import random
from getpass import getpass
from pkg_resources import resource_string, resource_filename
# Python 2/3 compat
from future.utils import iteritems  # pylint: disable=E0401
import aomi.exceptions


def my_version():
    """Return the version, checking both packaged and development locations"""
    if os.path.exists(resource_filename(__name__, 'version')):
        return resource_string(__name__, 'version')
    else:
        return open(os.path.join(os.path.dirname(__file__),
                                 "..", "version")).read()

VERSION = my_version()


def log(msg, opt):
    """Verbose messaging!"""
    if opt.verbose:
        print(msg, file=sys.stderr)


def warning(msg):
    """Print a warning message to stderr"""
    print("Warning {0}".format(msg), file=sys.stderr)


def abspath(raw):
    """Return what is hopefully a OS independent path."""
    path_bits = []
    if raw.find('/') != -1:
        path_bits = raw.split('/')
    elif raw.find('\\') != -1:
        path_bits = raw.split('\\')
    else:
        path_bits = [raw]

    return os.path.abspath(os.sep.join(path_bits))


def hard_path(path, prefix_dir):
    """Returns an absolute path to either the relative or absolute file."""
    relative = abspath("%s/%s" % (prefix_dir, path))
    if os.path.exists(relative):
        return relative

    return abspath(path)


def is_tagged(has_tags, required_tags):
    """Checks if tags match"""
    if len(required_tags) == 0 and len(has_tags) == 0:
        return True
    elif len(required_tags) == 0:
        return False
    else:
        found_tags = []
        for tag in required_tags:
            if tag in has_tags:
                found_tags.append(tag)

        return len(found_tags) == len(required_tags)


def cli_hash(list_of_kv):
    """Parse out a hash from a list of key=value strings"""
    ev_obj = {}
    for extra_var in list_of_kv:
        ev_list = extra_var.split('=')
        key = ev_list[0]
        val = '='.join(ev_list[1:])  # b64 and other side effects
        ev_obj[key] = val

    return ev_obj


def merge_dicts(dict_a, dict_b):
    """Deep merge of two dicts"""
    obj = {}
    for key, value in iteritems(dict_a):
        if key in dict_b:
            if isinstance(dict_b[key], dict):
                obj[key] = merge_dicts(value, dict_b.pop(key))
        else:
            obj[key] = value

    for key, value in iteritems(dict_b):
        obj[key] = value

    return obj


def get_tty_password(opt, confirm):
    """When returning a password from a TTY we assume a user
    is entering it on a keyboard so we ask for confirmation."""
    log("Reading password from TTY", opt)
    new_password = getpass('Enter Password: ', stream=sys.stderr)
    if not new_password:
        raise aomi.exceptions.AomiCommand("Must specify a password")

    if not confirm:
        return new_password

    confirm_password = getpass('Again, Please: ', stream=sys.stderr)
    if confirm_password != new_password:
        raise aomi.exceptions.AomiCommand("Passwords do not match")

    return new_password


def get_stdin_password(opt):
    """Returns a password from stdin, no confirmation neccesary"""
    log("Reading password from stdin", opt)
    return sys.stdin.readline().rstrip()


def get_password(opt, confirm=True):
    """Will return a password in an appropriate fashion"""
    if sys.stdin.isatty():
        return get_tty_password(opt, confirm)
    else:
        return get_stdin_password(opt)


def path_pieces(vault_path):
    """Will return a two part tuple comprising of the vault path
    and the key with in the stored object"""
    path_bits = vault_path.split('/')
    path = '/'.join(path_bits[0:len(path_bits) - 1])
    key = path_bits[len(path_bits) - 1]
    return path, key


def mount_for_path(path, client):
    """Returns the mountpoint for this path"""
    backend_data = client.list_secret_backends()['data']
    backends = [mnt for mnt in backend_data.keys()]
    path_bits = path.split('/')
    if len(path_bits) == 1:
        vault_path = "%s/" % path
        if vault_path in backends:
            return vault_path[0:len(vault_path) - 1]
    else:
        for i in range(1, len(path_bits) + 1):
            vault_path = "%s/" % '/'.join(path_bits[0:i])
            if vault_path in backends:
                return vault_path[0:len(vault_path) - 1]

    return None


def backend_type(path, client):
    """Returns the type of backend at the given mountpoint"""
    backends = client.list_secret_backends()['data']
    vault_path = "%s/" % path
    return backends[vault_path]['type']


def load_word_file(filename):
    """Loads a words file as a list of lines"""
    words_file = resource_filename(__name__, "words/%s" % filename)
    handle = open(words_file, 'r')
    words = handle.readlines()
    handle.close()
    return words


def choose_one(things):
    """Returns a random entry from a list of things"""
    choice = random.randint(0, len(things) - 1)
    return things[choice].strip()


def random_word():
    """Returns a random word string"""
    animal = choose_one(load_word_file("animals.txt"))
    academic = choose_one(load_word_file("academic.txt"))
    return "%s-%s" % (academic, animal)


def flatten(iterable):
    """Ensure we are returning an actual list, as that's all we
    are ever going to flatten within aomi"""
    return [x for x in actually_flatten(iterable)]


def actually_flatten(iterable):
    """Flatten iterables"""
    remainder = iter(iterable)
    while True:
        first = next(remainder)
        # Python 2/3 compat
        try:
            basestring
        except NameError:
            # Python 2/3 compat
            basestring = str  # pylint: disable=W0622
        if isinstance(first, collections.Iterable) and \
           not isinstance(first, basestring):
            remainder = IT.chain(first, remainder)
        else:
            yield first


def subdir_path(directory, relative):
    """Returns a file path relative to another path."""
    item_bits = directory.split(os.sep)
    relative_bits = relative.split(os.sep)
    for i in range(0, len(item_bits)):
        if i == len(relative_bits) - 1:
            return os.sep.join(item_bits[i:])
        else:
            if item_bits[i] != relative_bits[i]:
                return None

    return None



"""Hande various kinds of import/export of secrets"""
from __future__ import print_function
import os
import tempfile
import shutil
import time
import datetime
import zipfile

from aomi.helpers import warning, hard_path, log, \
    subdir_path
from aomi.vault import get_secretfile
from aomi.gpg import key_from_keybase, has_gpg_key, \
    import_gpg_key, encrypt, decrypt
from aomi.validation import sanitize_mount
import aomi.exceptions
from aomi.util import validate_entry


def from_keybase(username, opt):
    """Will attempt to retrieve a GPG public key from
    Keybase, importing if neccesary"""
    public_key = key_from_keybase(username)
    fingerprint = public_key['fingerprint'][-8:].upper().encode('ascii')
    key = public_key['bundle'].encode('ascii')
    if not has_gpg_key(fingerprint):
        log("Importing gpg key for %s" % username, opt)
        if not import_gpg_key(key):
            raise aomi.exceptions.KeybaseAPI("import key for %s" % username)

    return fingerprint


def grok_keys(config, opt):
    """Will retrieve a GPG key from either Keybase or GPG directly"""
    key_ids = []
    for key in config['pgp_keys']:
        if key.startswith('keybase:'):
            key_id = from_keybase(key[8:], opt)
            log("Encrypting for keybase user %s" % key[8:], opt)
        else:
            if not has_gpg_key(key):
                raise aomi.exceptions.GPG("Do not actually have key %s" % key)

            log("Encrypting for gpg id %s" % key, opt)
            key_id = key

        key_ids.append(key_id)

    return key_ids


def freeze_secret(src, dest, flav, tmp_dir, opt):
    """Copies a secret into a particular location"""
    src_file = hard_path(src, opt.secrets)
    dest_file = "%s/%s" % (tmp_dir, dest)
    dest_dir = os.path.dirname(dest_file)
    if not os.path.isdir(dest_dir):
        os.mkdir(dest_dir, 0o700)

    shutil.copy(src_file, dest_file)
    log("Froze %s %s" % (flav, src), opt)


def freeze_generic_file(secret, tmp_dir, opt):
    """Handles the potential validation of any frozen files"""
    path = "%s/%s" % (sanitize_mount(secret['mount']),
                      secret['path'])
    if not validate_entry(secret, path, opt):
        return

    for a_secret in secret['files']:
        sfile = a_secret['source']
        freeze_secret(sfile, sfile, 'file', tmp_dir, opt)


def freeze_aws_file(secret, tmp_dir, opt):
    """Handles the potential validation of any frozen AWS credentials"""
    path = "%s/config/root" % (sanitize_mount(secret['mount']))
    if not validate_entry(secret, path, opt):
        return

    sfile = secret['aws_file']
    freeze_secret(sfile, sfile, 'aws_file', tmp_dir, opt)


def freeze_var_file(secret, tmp_dir, opt):
    """Handles the potential validation of any frozen var file"""
    path = "%s/%s" % (sanitize_mount(secret['mount']), secret['path'])
    if not validate_entry(secret, path, opt):
        return

    sfile = secret['var_file']
    freeze_secret(sfile, sfile, 'var_file', tmp_dir, opt)


def freeze_files(config, tmp_dir, opt):
    """Copy files which are to be frozen to their temporary location"""
    for app in config.get('apps', []):
        freeze_secret(app['app_file'], app['app_file'], 'app', tmp_dir, opt)

    for user in config.get('users', []):
        pfile = user['password_file']
        freeze_secret(pfile, pfile, 'user', tmp_dir, opt)

    for secret in config.get('secrets', []):
        if 'var_file' in secret:
            freeze_var_file(secret, tmp_dir, opt)
        elif 'aws_file' in secret:
            freeze_aws_file(secret, tmp_dir, opt)
        elif 'files' in secret:
            freeze_generic_file(secret, tmp_dir, opt)

    for duo in config.get('duo', []):
        d_file = duo['creds']
        freeze_secret(d_file, d_file, 'duo', tmp_dir, opt)


def freeze_archive(tmp_dir, dest_prefix):
    """Generates a ZIP file of secrets"""
    zip_filename = "%s/aomi-blah.zip" % tmp_dir
    archive = zipfile.ZipFile(zip_filename, 'w')
    for root, _dirnames, filenames in os.walk(dest_prefix):
        for filename in filenames:
            relative_path = subdir_path(root, dest_prefix).split(os.sep)[1:]
            relative_path = os.sep.join(relative_path)
            archive.write("%s/%s" % (root, filename),
                          "%s/%s" % (relative_path, filename))

    archive.close()
    return zip_filename


def freeze_encrypt(dest_dir, zip_filename, config, opt):
    """Encrypts the zip file"""
    pgp_keys = grok_keys(config, opt)
    ice_handle = os.path.basename(os.path.dirname(opt.secretfile))
    timestamp = time.strftime("%H%M%S-%m-%d-%Y",
                              datetime.datetime.now().timetuple())
    ice_file = "%s/aomi-%s-%s.ice" % (dest_dir, ice_handle, timestamp)
    if not encrypt(zip_filename, ice_file, pgp_keys, opt):
        raise aomi.exceptions.GPG("Unable to encrypt zipfile")

    return ice_file


def freeze(dest_dir, opt):
    """Iterates over the Secretfile looking for secrets to freeze"""
    if not (os.path.exists(dest_dir) and
            os.path.isdir(dest_dir)):
        os.mkdir(dest_dir)

    tmp_dir = tempfile.mkdtemp('aomi-freeze')
    dest_prefix = "%s/dest" % tmp_dir
    os.mkdir(dest_prefix)
    config = get_secretfile(opt)

    freeze_files(config, dest_prefix, opt)
    zip_filename = freeze_archive(tmp_dir, dest_prefix)
    ice_file = freeze_encrypt(dest_dir, zip_filename, config, opt)
    shutil.rmtree(tmp_dir)
    log("Generated file is %s" % ice_file, opt)


def thaw_decrypt(src_file, tmp_dir, opt):
    """Decrypts the encrypted ice file"""

    if not os.path.isdir(opt.secrets):
        warning("Creating secret directory %s" % opt.secrets)
        os.mkdir(opt.secrets)

    zip_file = "%s/aomi.zip" % tmp_dir

    if not decrypt(src_file, zip_file, opt):
        raise aomi.exceptions.GPG("Unable to gpg")

    return zip_file


def thaw_secret(filename, tmp_dir, flav, opt):
    """Will perform some validation and copy a
    decrypted secret to it's final location"""
    src_file = "%s/%s" % (tmp_dir, filename)
    dest_file = "%s/%s" % (opt.secrets, filename)
    if not os.path.exists(src_file):
        raise aomi.exceptions.IceFile("%s file %s missing" % (flav, filename))

    shutil.copy(src_file, dest_file)
    log("Thawed %s %s" % (flav, filename), opt)


def thaw(src_file, opt):
    """Given the combination of a Secretfile and the output of
    a freeze operation, will restore secrets to usable locations"""
    if not os.path.exists(src_file):
        raise aomi.exceptions.AomiFile("%s does not exist" % src_file)

    tmp_dir = tempfile.mkdtemp('aomi-freeze')

    zip_file = thaw_decrypt(src_file, tmp_dir, opt)

    archive = zipfile.ZipFile(zip_file, 'r')
    for archive_file in archive.namelist():
        archive.extract(archive_file, tmp_dir)
        os.chmod("%s/%s" % (tmp_dir, archive_file), 0o640)
        log("Extracted %s from archive" % archive_file, opt)

    log("Thawing secrets into %s" % opt.secrets, opt)
    config = get_secretfile(opt)
    for app in config.get('apps', []):
        thaw_secret(app['app_file'], tmp_dir, 'App', opt)

    for user in config.get('users', []):
        thaw_secret(user['password_file'], tmp_dir, 'User', opt)

    for secret in config.get('secrets', []):
        if 'var_file' in secret:
            thaw_var_file(secret, tmp_dir, opt)
        elif 'aws_file' in secret:
            thaw_aws_file(secret, tmp_dir, opt)
        elif 'files' in secret:
            thaw_files(secret, tmp_dir, opt)

    for duo in config.get('duo', []):
        thaw_secret(duo['creds'], tmp_dir, 'DUO', opt)


def thaw_var_file(secret, tmp_dir, opt):
    """Thaw the contents of a var file"""
    path = "%s/%s" % (sanitize_mount(secret['mount']), secret['path'])
    if not validate_entry(secret, path, opt):
        return

    dest_file = "%s/%s" % (opt.secrets, secret['var_file'])
    var_file = os.path.basename(dest_file)
    src_file = "%s/%s" % (tmp_dir, var_file)
    if not os.path.exists(src_file):
        raise aomi.exceptions.IceFile("Var file %s missing" % var_file)

    shutil.copy(src_file, dest_file)
    log("Thawed var_file %s" % var_file, opt)


def thaw_aws_file(secret, tmp_dir, opt):
    """Thaw the contents of an AWS file"""
    path = "%s/config/root" % (sanitize_mount(secret['mount']))
    if not validate_entry(secret, path, opt):
        return

    dest_file = "%s/%s" % (opt.secrets, secret['aws_file'])
    aws_file = os.path.basename(dest_file)
    src_file = "%s/%s" % (tmp_dir, aws_file)
    if not os.path.exists(src_file):
        raise aomi.exceptions.IceFile("AWS file %s missing" % aws_file)

    shutil.copy(src_file, dest_file)
    log("Thawed aws_file %s" % aws_file, opt)


def thaw_files(secret, tmp_dir, opt):
    """Thaw some files"""
    for a_secret in secret['files']:
        path = "%s/%s" % (sanitize_mount(secret['mount']),
                          secret['path'])
        if not validate_entry(secret, path, opt):
            return

        filename = a_secret['source']
        dest_file = "%s/%s" % (opt.secrets, a_secret['source'])
        src_file = "%s/%s" % (tmp_dir, filename)
        if not os.path.exists(src_file):
            raise aomi.exceptions.IceFile("File %s missing" % filename)

        dest_dir = os.path.dirname(dest_file)
        if not os.path.isdir(dest_dir):
            os.mkdir(dest_dir, 0o700)

        shutil.copy(src_file, dest_file)
        log("Thawed file %s" % filename, opt)



"""Some validation helpers for aomi"""
from __future__ import print_function
import os
import platform
import stat
from aomi.helpers import abspath, is_tagged, log, subdir_path
import aomi.exceptions


def find_file(name, directory):
    """Searches up from a directory looking for a file"""
    path_bits = directory.split(os.sep)
    for i in range(0, len(path_bits) - 1):
        check_path = path_bits[0:len(path_bits) - i]
        check_file = "%s%s%s" % (os.sep.join(check_path), os.sep, name)
        if os.path.exists(check_file):
            return abspath(check_file)

    return None


def in_file(string, search_file):
    """Looks in a file for a string."""
    handle = open(search_file, 'r')
    for line in handle.readlines():
        if string in line:
            return True

    return False


def gitignore(opt):
    """Will check directories upwards from the Secretfile in order
    to ensure the gitignore file is set properly"""
    directory = os.path.dirname(abspath(opt.secretfile))
    gitignore_file = find_file('.gitignore', directory)
    if gitignore_file:
        secrets_path = subdir_path(abspath(opt.secrets), gitignore_file)
        if secrets_path:
            if not in_file(secrets_path, gitignore_file):
                e_msg = "The path %s was not found in %s" \
                        % (secrets_path, gitignore_file)
                raise aomi.exceptions.AomiFile(e_msg)
        else:
            log("Using a non-relative secret directory", opt)

    else:
        raise aomi.exceptions.AomiFile("You should really have a .gitignore")


def secret_file(filename):
    """Will check the permissions of things which really
    should be secret files"""
    filestat = os.stat(abspath(filename))
    if stat.S_ISREG(filestat.st_mode) == 0 and \
       stat.S_ISLNK(filestat.st_mode) == 0:
        e_msg = "Secret file %s must be a real file or symlink" % filename
        raise aomi.exceptions.AomiFile(e_msg)

    if platform.system() != "Windows":
        if filestat.st_mode & stat.S_IROTH or \
           filestat.st_mode & stat.S_IWOTH or \
           filestat.st_mode & stat.S_IWGRP:
            e_msg = "Secret file %s has too loose permissions" % filename
            raise aomi.exceptions.AomiFile(e_msg)


def validate_obj(keys, obj):
    """Super simple "object" validation."""
    msg = ''
    for k in keys:
        if isinstance(k, str):
            if k not in obj or len(obj[k]) == 0:
                if len(msg) > 0:
                    msg = "%s," % msg

                msg = "%s%s" % (msg, k)
        elif isinstance(k, list):
            found = False
            for k_a in k:
                if k_a in obj:
                    found = True

            if not found:
                if len(msg) > 0:
                    msg = "%s," % msg

                msg = "%s(%s" % (msg, ','.join(k))

    if len(msg) > 0:
        msg = "%s missing" % msg

    return msg


def var_file_obj(obj):
    """Does some validation around a var_file object"""
    check_obj(['var_file', 'mount', 'path'], 'var_file', obj)


def aws_file_obj(obj):
    """Does some validation around an aws_file object"""
    check_obj([['aws_file', 'aws'], 'mount'], 'aws_file', obj)


def aws_secret_obj(filename, obj):
    """Does some validation around AWS secrets"""
    check_obj(['access_key_id', 'secret_access_key'],
              "aws secret %s" % (filename),
              obj)


def aws_role_obj(obj):
    """Does some validation around an AWS role"""
    check_obj(['name'], 'aws role', obj)
    if obj.get('state', 'present') == 'present':
        check_obj([['policy', 'arn'], 'name'], 'aws role', obj)


def file_obj(obj):
    """Basic validation around file objects"""
    check_obj(['mount', 'path'], 'file element', obj)


def tag_check(obj, path, opt):
    """If we require tags, validate for that"""
    if not is_tagged(opt.tags, obj.get('tags', [])):
        log("Skipping %s as it does not have requested tags" %
            path, opt)
        return False
    else:
        return True


def specific_path_check(path, opt):
    """Will make checks against include/exclude to determine if we
    actually care about the path in question."""
    if len(opt.exclude) > 0:
        if path in opt.exclude:
            return False

    if len(opt.include) > 0:
        if path not in opt.include:
            return False

    return True


def check_obj(keys, name, obj):
    """Do basic validation on an object"""
    msg = validate_obj(keys, obj)
    if len(msg) > 0:
        raise aomi.exceptions.AomiData("object check : %s in %s" % (msg, name))


def policy_obj(obj):
    """Basic validation around policy objects"""
    state = obj.get('state', 'present')
    if state == 'present':
        check_obj(['name', 'file'], 'policy', obj)
    elif state == 'absent':
        check_obj(['name'], 'policy', obj)
    else:
        raise aomi.exceptions.AomiData("Invalid policy state: %s" % state)


def user_obj(obj):
    """Do basic validation on a user obj"""
    check_obj(['username', 'password_file', 'policies'],
              'user specification',
              obj)


def audit_log_obj(obj):
    """Do basic validation on an audit log object"""
    check_obj(['type'], 'audit log object', obj)
    if obj['type'] == 'file':
        check_obj(['file_path'], 'file audit log', obj)


def approle_obj(obj):
    """Do some basic approle validation"""
    check_obj(['name', 'policies'], 'app role', obj)


def generated_obj(obj):
    """Do some basic generated secret validation"""
    check_obj(['mount', 'path', 'keys'], 'generated secret object', obj)
    for key in obj['keys']:
        check_obj(['name', 'method'], 'generated secret entry', key)


def sanitize_mount(mount):
    """Returns a quote-unquote sanitized mount path"""
    sanitized_mount = mount
    if sanitized_mount.startswith('/'):
        sanitized_mount = sanitized_mount[1:]

    if sanitized_mount.endswith('/'):
        sanitized_mount = sanitized_mount[:-1]

    return sanitized_mount


def mount_obj(mount):
    """validates a mountpoint object"""
    check_obj(['path'], 'mount object', mount)


def duo_obj(obj):
    """Validates a duo obj"""
    check_obj(['host', 'creds', 'backend'], 'duo object', obj)
    if obj['backend'] != 'userpass':
        raise aomi.exceptions.AomiData('Invalid duo backend selected')
